import argparse
from os.path import dirname, join
from threading import Thread, Event
import sys

from bokeh.io import curdoc
from bokeh.layouts import row, column, widgetbox, gridplot
from bokeh.models import ColumnDataSource, Slider, Div, Button, FileInput 
from bokeh.events import Tap 

from .audio_widgets import Time_Waveform, Spectrum, CircularEq, WaveformEnvelope
from .audio_widgets import WaveformEnvelope, Centroid


from . import audio 
import IPython.display as ipd
import time


audioThread_ = None 
audio_thread_started_ = False
visualize_callback_active_ = False 
callback_id_ = None 

def update():
    """
    Periodic callback for updating all visualizations
    of all the audio widgets
    """
    
    # add the gain value of the slider to the data 
    #audio.data['gain'] = gain.value 

    # update the visualizers with the features extracted from
    # the audio analysis and playback thread
    for k, v in audio_widgets_.items():
        v.update(audio.data)

def create_audio_widgets(audio_file_name):
    """
    Create the audio widgets that will be used for
    the reactive visualizations.
    """
    
    audio_widgets = {} 
    audio_widgets['spectrum'] = Spectrum()
    audio_widgets['time_waveform'] = Time_Waveform()
    audio_widgets['waveform_envelope'] = WaveformEnvelope(audio_file_name, waveform_click_detected)
    audio_widgets['circulareq'] = CircularEq()
    audio_widgets['centroid'] = Centroid() 
    return audio_widgets 



def file_input_handler(attr, old, new):
    """
    Handler that gets called when the user chooses a file
    Args:
        attr: 
        old: the old value of the file
        new: the new value of the file 
    """ 
    
    print('File input handler')
    print(file_input.filename)
    global audio_file_name_ 
    audio_file_name_  = file_input.filename
    
    
def play_handler():
    """
    Gets called when the play button is pressed
    and starts the audio playback
    """
    
    global callback_id_
    global audio_playing_
    global audio_thread_started_
    global visualize_callback_active_
    
    audio_close.clear()
    audio_playing_ = True
    if not audio_thread_started_:
        start_audio_thread(audio_file_name_)
        audio_thread_started_ = True

    if args.playback_mode == "html":
        s1 = "document.getElementById('myaudio').play();"
        ipd.display(ipd.Javascript(s1))
    audio_play.set()

    if (visualize_callback_active_ is False):
        callback_id_ = curdoc().add_periodic_callback(update, 60)
    visualize_callback_active_ = True


def pause_handler():
    """
    Pause the audio playback
    """
    global audio_playing_
    audio_playing_ = False
    if args.playback_mode == "html":
        s1 = "document.getElementById('myaudio').pause();"
        ipd.display(ipd.Javascript(s1))
    audio_play.clear()


def close_handler():
    """
    Stop the audio thread and the visualization callback
    """

    curdoc().remove_periodic_callback(callback_id_)
    if args.playback_mode == "html":
        s1 = "document.getElementById('myaudio').pause();"
        ipd.display(ipd.Javascript(s1))
    audio_play.clear()
    audio_close.set()
    global audioThread_
    if audioThread_:
        audioThread_.join
    global audio_thread_started_
    audio_thread_started_ = False
    audio_play.set()


def waveform_click_detected(event):
    """
    Seek a particular audio location based on event

    Args:
    event: contains the x,y coordinates of the mouse click in data
    coordinates
    """

    if args.playback_mode == "html":
        s1 = "document.getElementById('myaudio').currentTime = "
        + str(event.x) + ";"
        ipd.display(ipd.Javascript(s1))
    audio.set_current_time(event.x)
    audio_seek.set()


def start_audio_thread(audio_file_name):
    """
    Start the audio thread
    Args:
    audio_file_name (str): the audio file from which to read the samples
    """
    global audio_thread_started_
    audio_thread_started_ = True 
    global audioThread_
    audioThread_ = Thread(target=audio.update_audio_data,
                          args=(audio_file_name,
                                args.playback_mode,
                                audio_play,
                                audio_close,
                                audio_seek))
    audioThread_.setDaemon(True)
    audioThread_.start()


audio_playing_ = False
parser = argparse.ArgumentParser()
parser.add_argument("audio_file_name",
                    type=str,
                    help="the audio file name"
                    )
parser.add_argument("playback_mode",
                    type=str,
                    choices=['pyaudio', 'html'],
                    default='pyaudio',
                    help="The playback mode. One of ['pyaudio', 'html']"
                    )
parser.add_argument("widgets",
                    nargs='+',
                    default=[],
                    choices=['time_waveform',
                             'spectrum',
                             'circulareq',
                             'waveform_envelope',
                             'centroid'],
                    help="The audio widget names."
                    )
args = parser.parse_args()
audio_file_name_ = args.audio_file_name

# preload the audio for html playback
if args.playback_mode == "html":
    s = '<audio id="myaudio" src="'
    + audio_file_name_ + '" preload="auto"></audio>'
    ipd.display(ipd.HTML(s))


# create visualizers
audio_widgets_ = create_audio_widgets(audio_file_name_)

# start audio thread - audio_play event controls play/pause
audio_play = Event()
audio_close = Event()
audio_seek = Event()


control_grid = []
# setup some sliders for controlling the widgets
max_freq = 22050
#freq = Slider(start=1, end=max_freq, value=max_freq, step=1, title="Frequency")
#gain = Slider(start=1, end=20, value=1, step=1, title="Gain")

#sound_control = [gain, freq]
#control_grid.append(sound_control)

# setup playback widgets
play_button = Button(label="Play", button_type="success")
pause_button = Button(label="Pause", button_type="success")
close_button = Button(label="Close", button_type="success")

playback_control = [play_button, pause_button, close_button]
play_button.on_click(play_handler)
pause_button.on_click(pause_handler)
close_button.on_click(close_handler)
control_grid.append(playback_control)


file_input_ = FileInput(accept=".wav")
file_input_.on_change('filename', file_input_handler)


# setup the document
filename = join(dirname(__file__), "description.html")
desc = Div(text=open(filename).read(), render_as_text=False, width=1000)
curdoc().add_root(desc)
curdoc().add_root(gridplot(control_grid))

# make a column of desired widgets
plots = [audio_widgets_[x].get_plot() for x in args.widgets]
#for p in plots: 
#    p.on_event(Tap, waveform_click_detected)

plot_column = column(plots)
curdoc().add_root(plot_column)
curdoc().add_root(file_input_)



import os
import time
from time import sleep
import numpy as np
from scipy import fft 
import soundfile as sf
from scipy.integrate import simps
from .audio_processors import SpectrumAudio, SpectrumBins, SpectralCentroid

# These are the values extracted on every block
# of audio that is being processed 
data = {'time': 0.0,
        'signal': None,
        'gain': None}


# information about the audio file 
audio_info = {'duration': None,
              'samplerate': None,
              'blockSize': None}


ctime_ = 0.0 
    

def update_audio_data(fname, playback_mode, audio_play,
                      audio_close, audio_seek):

    """ Thread that reads and processes the audio data using audio processors 

    Args:
        fname (str) : the file from which the audio is read 
        playback_mode (str): pyaudio or html  
        playback_mode (str): pyaudio or html
        audio_play (Event): event for starting/pausing playback
        audio_close(Event): event for stopping the audio thread
        audio_seek(Event): event for audio seeking 
    """
 
    
    k = 0
    pyaudio_available = False

    # initialize stuff at first iteration 
    if k == 0:
        current_time = data['time']
        blockSize = 2048
        circBins = 16

        # open sound file and get info about it 
        isf = sf.SoundFile(fname, 'r')
        audio_info['duration'] = sf.info(fname).duration
        audio_info['samplerate'] = sf.info(fname).samplerate
        audio_info['blockSize'] = blockSize
        block_duration = float(blockSize) / audio_info['samplerate']
        
        # allocate arrays 
        signal = np.zeros(blockSize)
        html_data = signal.astype(np.float32)
        len_samples = len(isf)
        len_blocks = int(len_samples / blockSize)

        # initialize data dictionary 
        data['signal'] = signal
        data['spectrum'] = np.zeros(blockSize)
        data['bins'] = np.zeros(circBins) 
        data['centroid_track'] = np.zeros(len_blocks)
        data['scentroid_track'] = np.zeros(len_blocks)        
        
        # initialize processors 
        spectrum_processor = SpectrumAudio(blockSize)
        spectrum_bins_processor = SpectrumBins(circBins)
        centroid_processor = SpectralCentroid(len_blocks)
        
        # initialize pyaudio 
        try:
            import pyaudio
            # setup audio playback stream 
            pa = pyaudio.PyAudio()
            stream = pa.open(
                format=pyaudio.paFloat32,
                channels=1,
                rate=int(audio_info['samplerate']),
                input=False,
                output=True,
                frames_per_buffer=audio_info['blockSize']
            )
            pyaudio_available = True
        except:
            pyaudio_available = False

    while True:
        current_time = data['time']
        try:
            t0 = time.perf_counter()
            
            # wait on event to start audio playback 
            audio_play.wait()

            # read from input sound file 
            sfdata = isf.read(blockSize, always_2d=True).astype(np.float32)
            # convert to mono 
            sfdata =  sfdata.sum(axis=1) / 2

            # seek if needed 
            if audio_seek and audio_seek.isSet():
                set_time = ctime_
                set_frame = int(set_time*audio_info['samplerate'])
                isf.seek(set_frame)
                current_time = set_time
                audio_seek.clear()

            # write audio to stream
            if playback_mode == "html":
                stream.write(html_data.tostring())
            elif playback_mode == "pyaudio":
                if pyaudio_available:                    
                    stream.write(sfdata.tostring())
            else:
                print('playback_mode', playback_mode)
                print("Unsupported playback_mode")

            # apply the different processors
            # notice that the spectrum_bins_processor
            # needs to come after the spectrum processor
            # as it requires the spectrum to be present
            # in data. Essentially each processor adds
            # new information to the data dictionary 
            spectrum_processor.process(data)
            spectrum_bins_processor.process(data)
            centroid_processor.process(data)
            
            # advance time
            current_time += block_duration
            t1 = time.perf_counter()

            data['time'] = current_time                
            data['signal'] = sfdata

            # sleep for approximately the right number
            # of time for processing a block
            # in order for the html/javascript playback
            # to work on remote notebooks
            
            if not pyaudio_available:
                if block_duration > (t1 - t0):
                    sleep(block_duration - (t1 - t0))
                else:
                    sleep(t1 - t0)

            # deal with end of file with looping to the start 
            k = k + 1
            if k >= len_blocks:
                print("End of file")
                set_current_time(0.0)
                k = 0
                isf.seek(0)

            # deal with closing/terminating the audio thread 
            if audio_close.isSet():
                print("Terminating audio analysis/playback")
                set_current_time(0.0)
                k = 0
                isf.seek(0)                                
                return 
        except:
            continue


        
def get_current_time():
    """ Return the current time in seconds  
    """
    return data['time']

def set_current_time(t):
    """ Set the current time
    Args:
       t (float): the time in seconds 
    """
    
    data['time'] = t
    global ctime_ 
    ctime_ =  t





from math import ceil
import numpy as np
import soundfile as sf
from bokeh.models import ColumnDataSource, LabelSet
from bokeh.plotting import figure
from bokeh.events import Tap
from .audio import *


class WaveformEnvelope:

    def __init__(self, fname, tap_function):
        plotargs = dict(tools="", toolbar_location=None, outline_line_color='#595959')

        self.fname = fname
        self.tap_function = tap_function 
        time = [0.0, 1.0] 
        self.signal_plot = figure(plot_width=1000, plot_height=350,
                                  title="Waveform Envelope",
                                  x_range=[0, time[-1] + 30],
                                  y_range=[-1.25, 1.25],
                                  x_axis_label='Time(secs)',
                                  y_axis_label='Amplitude', **plotargs)
        self.signal_plot.background_fill_color = "#eaeaea"


        
        try:        
            audio, srate = sf.read(fname, always_2d=True)
        except RuntimeError:
            print('Problem opening ', fname)
            return

        # convert to mono 
        audio =  audio.sum(axis=1) / 2

        max_envelope = self.process(audio, 4096, block_process=self.max_absolute)
        time = np.linspace(0, len(audio) / srate, num=len(max_envelope))    



        self.signal_source_pos = ColumnDataSource(data=dict(t=[], y=[]))
        self.signal_source_neg = ColumnDataSource(data=dict(t=[], y=[]))

        self.signal_plot.vbar(x=time, bottom=-max_envelope,
                              top=max_envelope,
                              width=1.0 / len(time),
                              fill_color='gray',
                              fill_alpha=0.5)

        self.duration = len(audio) / srate
        self.length = len(max_envelope)        

        t = np.linspace(0, self.duration, self.length)
        y = np.zeros(len(t))

        #self.signal_plot.line(x="t", y="y", line_color="green",
        #                      source=self.signal_source_pos, line_width=1.0)
        #self.signal_plot.line(x="t", y="y", line_color="green",
        #                      source=self.signal_source_neg, line_width=1.0)
        self.signal_plot.on_event(Tap, self.tap_detected)
        y[0] = 1.25
        
        self.signal_plot.vbar(x="t", top="y", width = 20.0 / len(time) ,
                              color = 'red', fill_color = 'red',fill_alpha=1.0, 
                              source=self.signal_source_pos)


        self.signal_plot.vbar(x="t", top="y", width = 20.0 / len(time),
                              color = 'red', fill_color = 'red', fill_alpha=1.0, 
                              source=self.signal_source_neg)
        

    def tap_detected(self, event):
        self.tap_function(event) 
        

    def max_absolute(self, block):
        return np.max(np.abs(block))

    def process(self, signal, hopSize, block_process):
        winSize = hopSize
        offsets = np.arange(0, len(signal), hopSize)
        amp_track = np.zeros(len(offsets))
        for (m, o) in enumerate(offsets):
            frame = signal[o:o + winSize]
            amp_track[m] = block_process(frame)
        return amp_track

    def update(self, data):
        time = data['time']
        
        t = np.linspace(0, self.duration, self.length)
        y = np.zeros(len(t))
        if time is None:
            pass
        else:
            position = (time / self.duration) * self.length
            y[int(position)] = 1.25
            self.signal_plot.title.text = "Waveform Envelope - " + self.fname  + '\t' + str(time)

            self.signal_source_pos.data = dict(t=t, y=y)
            self.signal_source_neg.data = dict(t=t, y=-y)




            
            
    def get_plot(self):
        return self.signal_plot


class Time_Waveform:
    def __init__(self, time_range=46):
        plotargs = dict(tools="", toolbar_location=None,
                        outline_line_color='#595959')
        self.signal_source = ColumnDataSource(data=dict(t=[], y=[]))
        self.signal_plot = figure(plot_width=600, plot_height=200,
                                  title="Signal",
                                  x_range=[0, time_range],
                                  y_range=[-0.8, 0.8],
                                  x_axis_label='Time(msec)',
                                  y_axis_label='Amplitude', **plotargs)
        self.signal_plot.background_fill_color = "#eaeaea"
        self.signal_plot.line(x="t", y="y", line_color="#024768",
                              source=self.signal_source)

    def update(self, data):
        
        signal = data['signal']
        time = data['time']
        gain = data['gain']

        if signal is None:
            pass 
        else:
            t = np.linspace(0, 100, len(signal))            
            self.signal_source.data = dict(t=t, y=signal)
            self.signal_plot.title.text = "Time Domain Waveform\t" + str(time)

    def get_plot(self):
        return self.signal_plot


class Spectrum:
    def __init__(self):
        plotargs = dict(tools="", toolbar_location=None,
                        outline_line_color='#595959')
        max_freq_khz = 22050 * 0.1
        self.spectrum_source = ColumnDataSource(data=dict(f=[], y=[]))
        self.spectrum_plot = figure(plot_width=600, plot_height=200,
                                    title="Power Spectrum",
                                    y_range=[10 ** (-4), 10 ** 3],
                                    x_range=[0, max_freq_khz],
                                    y_axis_type="log", **plotargs)
        self.spectrum_plot.background_fill_color = "#eaeaea"
        self.spectrum_plot.line(x="f", y="y", line_color="#024768",
                                source=self.spectrum_source)

    def update(self, data):
        if 'spectrum' in data: 
            spectrum = data['spectrum']
        else:
            return 
            
        time = data['time']
        max_freq_khz = 22050 * 0.1
        if spectrum is None:
            pass
        else:
            f = np.linspace(0, max_freq_khz, len(spectrum))
            self.spectrum_source.data = dict(f=f, y=spectrum)
        # spectrum_plot.x_range.end = freq.value*0.001

    def get_plot(self):
        return self.spectrum_plot


class CircularEq:
    def __init__(self):
        num_bins = 16
        eq_clamp = 20
        plotargs = dict(tools="", toolbar_location=None,
                        outline_line_color='#595959')
        self.eq_range = np.arange(eq_clamp, dtype=np.float64)
        eq_angle = 2 * np.pi / num_bins

        eq_data = dict(
            inner=np.tile(self.eq_range + 2, num_bins),
            outer=np.tile(self.eq_range + 2.95, num_bins),
            start=np.hstack([np.ones_like(self.eq_range) * eq_angle * (i + 0.05) for i in range(num_bins)]),
            end=np.hstack([np.ones_like(self.eq_range) * eq_angle * (i + 0.95) for i in range(num_bins)]),
            alpha=np.tile(np.zeros_like(self.eq_range), num_bins),
        )
        self.eq_source = ColumnDataSource(data=eq_data)
        self.eq_plot = figure(plot_width=400, plot_height=400,
                              x_axis_type=None, y_axis_type=None,
                              x_range=[-20, 20], y_range=[-20, 20], **plotargs)
        self.eq_plot.background_fill_color = "#eaeaea"
        self.eq_plot.annular_wedge(x=0, y=0, fill_color="#024768",
                                   fill_alpha="alpha",
                                   line_color=None, inner_radius="inner",
                                   outer_radius="outer", start_angle="start",
                                   end_angle="end",
                                   source=self.eq_source)

    def update(self, data):
        if "bins" in data: 
            bins = data['bins']
        else:
            return 
        time = data['time']
        alphas = []
        if (bins is None):
            pass 
        else: 
            for x in bins:
                a = np.zeros_like(self.eq_range)
                N = int(ceil(x))
                a[:N] = (1 - self.eq_range[:N] * 0.05)
                alphas.append(a)
            self.eq_source.data['alpha'] = np.hstack(alphas)

    def get_plot(self):
        return self.eq_plot




class Centroid:
    def __init__(self):
        plotargs = dict(tools="", toolbar_location=None,
                        outline_line_color='#595959')

        self.centroid_source = ColumnDataSource(data=dict(t=[], y=[]))
        self.scentroid_source = ColumnDataSource(data=dict(t=[], y=[]))
        self.centroid_plot = figure(plot_width=800, plot_height=200,
                                    title="Centroid", x_range=[0, 100],
                                    y_range=[0, 200], **plotargs)
        self.centroid_plot.background_fill_color = "#eaeaea"
        self.centroid_plot.line(x="t", y="y", line_color="blue",
                                source=self.centroid_source)
        self.centroid_plot.line(x="t", y="y", line_color="red",
                                source=self.scentroid_source, line_width=3)

        self.scentroid_plot = figure(plot_width=800, plot_height=200,
                                     title="Smoothed Centroid (mean of previous 20 values)", x_range=[0, 100],
                                     y_range=[0, 200], **plotargs)
        self.scentroid_plot.background_fill_color = "#eaeaea"
        self.scentroid_plot.line(x="t", y="y", line_color="red",
                                 source=self.scentroid_source)

    def update(self, data):
        centroid_track = data['centroid_track']
        scentroid_track = data['scentroid_track']

        t = np.linspace(0, 100, len(centroid_track))
        self.centroid_source.data = dict(t=t, y=centroid_track)
        self.scentroid_source.data = dict(t=t, y=scentroid_track)

    def get_plot(self):
        return self.centroid_plot





import numpy as np
from scipy import fft
from scipy.integrate import simps


class SpectrumAudio:

    name = 'SpectrumAudio' 
    def __init__(self, blockSize):
        """
        Process audio input and return the corresponding power spectrum 
        Args:
            blockSize (int): the size in samples of the block to be processed
        """
        self.blockSize = blockSize 
        
    def process(self, data):
        """
        Process audio input and add the corresponding power spectrum
        to the data dictionary 
        
        Args:
            data (ndarray): the data containing the audio samples
        """
        
        signal = data['signal'] 
        cspectrum = fft.fft(signal)
        spectrum = abs(cspectrum)[:int(self.blockSize / 2)]
        power = spectrum ** 2
        data['spectrum'] = power



class SpectralCentroid:
    name = 'SpectralCentroid' 
    def __init__(self, len_blocks):
        """
        Process a spectrum and returns the corresponding centroid 
        """

        self.len_blocks = len_blocks
        self.centroid_track = np.zeros(len_blocks+1)
        self.scentroid_track = np.zeros(len_blocks+1)
        self.k = 0 
        
    def process(self, data):
        """
        Process spectrum input and add the centroid to
        the data dictionary 
        Args:
            data (ndarray): the data containing the audio samples
        """
        spectrum = data['spectrum']
        
        if (np.sum(spectrum) != 0.0):
            norm_spectrum = spectrum / np.sum(spectrum)
        else:
            norm_spectrum = spectrum 
        # centroid extraction
        fbins = np.arange(0.0, float(len(norm_spectrum)))
        centroid = np.dot(fbins, norm_spectrum)
        data['centroid_track'][self.k] = centroid
        
        if (self.k > 20):
            data['scentroid_track'][self.k] = np.sum(data['centroid_track'][self.k - 20:self.k]) / 20

        self.k += 1




class SpectrumBins:
    name = 'SpectrumBins' 
    def __init__(self, circBins):
        """
        Process the spectrum and a smaller number of bins summarizing it  
        Args:
            bins (int): number of bins for circular equalizer
        """
        
        self.circBins = circBins
        
    def process(self, data):
        """
        Process the spectrum input and add the circular bins to
        the data dictionary 
        Args:
            data (ndarray): the data containing the audio samples
        """
        
        spectrum = data['spectrum']
        bins = simps(np.split(spectrum, self.circBins))
        data['bins'] = bins 

        



import json
import re

import ipykernel
import requests
import soundfile as sf
from bokeh.application import Application
from bokeh.resources import INLINE
from bokeh.application.handlers import DirectoryHandler
from bokeh.io import show, output_notebook
from notebook import notebookapp
from requests.compat import urljoin
import tempfile


def get_notebook_url():
    
    """Tries to find out what the notebook url is
    
    Returns:
        a string with the notebook url 
    """

    servers = list(notebookapp.list_running_servers())

    # find kernel id 
    kernel_id = re.search('kernel-(.*).json',
                          ipykernel.connect.get_connection_file()).group(1)

    # look through the servers and try to match kernel id 
    notebook_url = ""
    for ss in servers:
        response = requests.get(urljoin(ss['url'], 'api/sessions'),
                                params={'token': ss.get('token', '')})
        kernels = json.loads(response.text)
        for nn in kernels:
            if 'kernel' in nn:
                if nn['kernel']['id'] == kernel_id:
                    notebook_url = ss['url']
    return notebook_url[:-1]


def render(audio_input, widgets=[], playback_mode="pyaudio", notebook_url=None):
 
    audio_fname = audio_input
    if type(audio_input) is tuple:     # audio data input 
        audio, samplerate = audio_input
        temp_file_name = tempfile.NamedTemporaryFile(dir='/tmp',
                                              suffix='.wav',
                                              delete=False)
        # write a temporary file for the audio data 
        sf.write(temp_file_name, audio, samplerate)        
        audio_fname = temp_file_name.name 

    # set up the command line arguments passed to the application 
    argv = [audio_fname]
    argv = argv + [playback_mode] + widgets
    
    # create a Bokeh application from the audio_widgets directory
    handler = DirectoryHandler(filename='audio_widgets', argv=argv)
    app = Application(handler)

    # show the application in the notebook 
    if notebook_url is None:
        notebook_url = get_notebook_url()
        print(notebook_url)
    output_notebook(resources=INLINE)
    show(app, notebook_url=notebook_url)


# short-hand calls for specific widgets 
    
def spectrum(audio_input, playback_mode = 'pyaudio', notebook_url=None):
    """ Spectrum audio widget
    """
    render(audio_input, ['spectrum'], playback_mode, notebook_url)


def time_waveform(audio_input, playback_mode = 'pyaudio', notebook_url=None):
    """ Time domain waveform 
    """
    render(audio_input, ['time_waveform'], playback_mode, notebook_url)

    
def waveform_envelope(audio_input, playback_mode = 'pyaudio', notebook_url=None):
    """ Waveform envelope 
    """
    render(audio_input, ['waveform_envelope'], playback_mode, notebook_url)

def circulareq(audio_input, playback_mode = 'pyaudio', notebook_url=None):

    """ Circular equalizer 
    """
    
    render(audio_input, ['circulareq'], playback_mode, notebook_url)


def centroid(audio_input, playback_mode = 'pyaudio', notebook_url=None):
    """ Spectrum audio widget
    """
    render(audio_input, ['centroid'], playback_mode, notebook_url)
    
